/*
 * filename: src/aofserv.c
 * author: Pablo Yanez Trujillo <yanezp@informatik.uni-freiburg.de>
 *
 */

#include <stdio.h>
#include <sqlite3.h>
#include <stdlib.h>
#include <config.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <syslog.h>

#define PORT 900
#define MAX_LINE 4096
#define PIDFILE "/var/run/aofserv.pid"

#include "conf_parse.h"

/* SELECT I.name,I.desc,I.type,I.year,C.name FROM image I JOIN category C ON C.id=I.cat ORDER BY I.cat */

/**
 * Protocol aofimg (TCP on Port 900)
 *
 * 1. Server listens on Port 900 (TCP)
 * 2. After accepting, client sends category name and a new-line.
 *    If all categories are needed, then client sends __all__\n
 * 3. Server sends 1 if it is able to send data, 0 otherwise.
 * 4. Server sends then number of images and then writes a new-line
 * 5. Server sends the data in the following order:
 *    1. Image name
 *    2. Image description
 *    3. Image type
 *    4. Image date
 *    5. Image category
 *
 *    The server sends first the length of the data and writes a new-line. Afterwards the server
 *    sends the data and continues this procedure until the connection socket is closed.
 *
 */

static int writeSQL(void *con, int argc, char **argv, char **azColName) 
{
	char line[MAX_LINE];
	int i, *connfd = con, datalen;

	for(i=0; i < argc; ++i)
	{
		datalen = strlen(argv[i]);
		sprintf(line, "%d\n", datalen);

		write(*connfd, line, strlen(line));
		write(*connfd, argv[i], datalen);
	}

	return 0;
}

static int numOfQueries(void *count, int argc, char **argv, char **azColName)
{
	int *num = count;

	*num = atoi(argv[0]);

	return 0;
}

int daemonize(void)
{
        pid_t child;

        if((child = fork()) == -1)
        {
                syslog(LOG_ERR, "cannot create a child process");
                fprintf(stderr, "cannot create a child process");

                return 0;
        }

        if(child)
                exit(EXIT_SUCCESS); /* terminate parent process and let
                                       child work as session leader */

        setsid();

        chdir("/");

        umask(0);

        return 1;
}

int listening(void)
{
	int sock, i;
	char msg[MAX_LINE];
	struct sockaddr_in address;

	if( (sock = socket(PF_INET, SOCK_STREAM, 0)) < 0)
	{
		syslog(LOG_ERR, "Cannot create a socket");
		fprintf(stderr, "Cannot create a socket\n");
		return -1;
	}

	i = 1;

	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i));

	address.sin_family = AF_INET;
	address.sin_port = htons(PORT);
	memset(&address.sin_addr, 0, sizeof(address.sin_addr));

	if(bind(sock, (struct sockaddr*) &address, sizeof(address)))
	{
		sprintf(msg, "Cannot bind on port %d", PORT);
		syslog(LOG_ERR, msg);
		fprintf(stderr, "%s\n", msg);
		close(sock);
		return -1;
	}

	if(listen(sock, 5))
	{
		sprintf(msg, "Cannot listen on port %d", PORT);
		syslog(LOG_ERR, msg);
		fprintf(stderr, "%s\n", msg);
		close(sock);
		return -1;
	}

	return sock;
}

void cleanup(int signr)
{
        if(unlink(PIDFILE))
                syslog(LOG_ERR, "Pid file %s can not be removed", PIDFILE);

        syslog(LOG_INFO, "server shutting down");

        closelog();

        exit(EXIT_SUCCESS); 
}

void connecting(int sock, const char* db)
{
	sqlite3 *sql_db;
	int count;
	char *err = NULL, category[MAX_LINE], sql[MAX_LINE], tmp[MAX_LINE];
	int connfd, rc, n, j;
	struct sockaddr_in address;
	size_t addrlength = sizeof(struct sockaddr_in);

	while( ((connfd = accept(sock, (struct sockaddr*) &address, &addrlength)) >= 0) )
	{
		syslog(LOG_INFO, "Accepting connection from %s", inet_ntoa(address.sin_addr));

		n = j = 0;

		while ( (n = read(connfd, &category[j], 1)) > 0)
		{
			if(category[j] == '\n' || category[j] == '\r')
			{
				category[j] = 0;
				break;
			}
			j++;
		}

		if(n < 0)
		{
			syslog(LOG_ERR, "Cannot read from socket: %s", strerror(errno));
			close(connfd);
			continue;
		}

		tmp[0] = 0;

		if(strcmp(category, "__all__"))
			sprintf(tmp, "AND C.name LIKE '%s'", category);

		sprintf(sql, "SELECT COUNT(I.name) FROM image I JOIN category C ON C.id=I.cat %s ORDER BY I.cat", tmp);

		rc = sqlite3_open(db, &sql_db);

		if(rc)
		{
			syslog(LOG_ERR, "Can't open database: %s", sqlite3_errmsg(sql_db));
			sqlite3_close(sql_db);
			write(connfd, "0", 1);
			close(connfd);
			continue;
		}

		count = 0;

		rc = sqlite3_exec(sql_db, sql, numOfQueries, &count, &err);

		if( rc!=SQLITE_OK )
		{
			syslog(LOG_ERR, "SQL-Lite error: %s", err);
			sqlite3_free(err);
			write(connfd, "0", 1);
			close(connfd);
			continue;
		}

		sprintf(sql, "1%d\n", count);

		write(connfd, sql, strlen(sql));

		sprintf(sql, "SELECT I.name,I.desc,I.type,I.year,C.name FROM image I JOIN category C ON C.id=I.cat %s ORDER BY I.cat", tmp);

		rc = sqlite3_exec(sql_db, sql, writeSQL, &connfd, &err);
		
		if( rc!=SQLITE_OK )
		{
			syslog(LOG_ERR, "SQL-Lite error: %s", err);
			sqlite3_free(err);
		}

		sqlite3_close(sql_db);
		close(connfd);
	}

	if(sock != -1)
		close(sock);
}


int main(void)
{
	char db[MAX_LINE];
	int sock;
	struct CONF conf;
	FILE* fp;

	if(getuid())
	{
		fprintf(stderr, "Only root can start sc-cron\n");
		return EXIT_FAILURE;
	}

	conf.db = NULL;
	conf.uid = -1;
	conf.gid = -1;

	if(!getconf(&conf))
	{
		free(conf.db);
		printf("error while reading config file\n");
		return EXIT_FAILURE;
	}

	strcpy(db, conf.db);
	free(conf.db);

	openlog("aofserv", LOG_PID, LOG_DAEMON);
	syslog(LOG_INFO, "staring server");

	if(signal(SIGINT, cleanup) == SIG_ERR)
        {
                syslog(LOG_ALERT, "The signal SIGINT could not be set");
                return EXIT_FAILURE;
        }

	if(!daemonize())
	{
		fprintf(stderr, "sc-cron cannot be start as a daemon\n");
		syslog(LOG_ERR, "sc-cron cannot be start as a daemon");
		return EXIT_FAILURE;
	}

	fp = fopen(PIDFILE, "w");

	if(!fp)
	{
		syslog(LOG_ALERT, "%s cannot be written: %m", PIDFILE);
		return EXIT_FAILURE;
	}

	fprintf(fp, "%d\n", getpid());
	fclose(fp);

	sock = listening();

	if(sock == -1)
		return EXIT_FAILURE;

	connecting(sock, db);

	closelog();

	close(sock);

	return EXIT_SUCCESS;
}
