/*
 * filename: src/install.c
 * author: Pablo Yanez Trujillo <yanezp@informatik.uni-freiburg.de>
 */

#include <stdio.h>
#include <sqlite3.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

#include "install.h"

static int callback(void *NotUsed, int argc, char **argv, char **azColName){
	/* this function does anything. Avoid warnings */
	NotUsed = NotUsed;
	argc = argc;
	argv = argv;
	azColName = azColName;
	return 0;
}

static int change_rights(const char* path, int uid, int gid, int dir)
{
	int rc;

	rc = chown(path, uid, gid);
        if(rc)
        {
                fprintf(stderr, "Could not change user and group id of `%s'.\n"
                                "Try to do it by yourself, otherwise you won't be\n"
                                "able to write in the databse.\n", path);
                return 0;
        }

        rc = chmod(path, S_IRUSR | S_IWUSR | (dir ? S_IXUSR : 0));
        if(rc)
        {
                fprintf(stderr, "Could not change the rights of `%s'.\n"
                                "Try to do it by yourself, otherwise unauthorized\n"
                                "users could write in and/or read from the database\n", path);
                return 0;
        }

	return 1;
}

int db_install(const char* path, int uid, int gid)
{
	const char* sql_statements[] = { 
"CREATE TABLE category (\n\
  id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,\n\
  subid INTEGER NOT NULL DEFAULT 0,\n\
  name TEXT NOT NULL,\n\
  UNIQUE(name))",

"CREATE TABLE image (\n\
  id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,\n\
  cat INTEGER NOT NULL,\n\
  name TEXT NOT NULL,\n\
  desc TEXT NOT NULL,\n\
  longdesc TEXT NOT NULL,\n\
  language TEXT NOT NULL,\n\
  type TEXT NOT NULL DEFAULT 'CD',\n\
  year TEXT NOT NULL,\n\
  author TEXT NOT NULL,\n\
  path TEXT NOT NULL)" };
	
	FILE* buffer;
	sqlite3 *db;
	char* dberror = NULL, *dir;
	int rc, i;


	if(!path || uid < 0 || gid < 0) return 0;

	buffer = fopen(path, "r");

	if(buffer)
	{
		fclose(buffer);
		fprintf(stderr, "Database file already exists. Cannot install a new databse on it\n"
				"without deleting the whole data. Please remove the file by yourself.\n");
		return 0;
	}

	rc = sqlite3_open(path, &db);

	switch(rc)
	{
		case SQLITE_OK:
			break;
		case SQLITE_CANTOPEN:
			fprintf(stderr, "Cannot create database file `%s'\n"
					"Take a look at /etc/aof-db.conf and create the needed directories.\n", path);
			sqlite3_close(db);
			return 0;
		default:
			fprintf(stderr, "Unexpected error while creating the database: %s\n", sqlite3_errmsg(db));
			sqlite3_close(db);
			return 0;
	}

	for(i = 0; i < 2; ++i)
	{
		rc = sqlite3_exec(db, sql_statements[i], callback, 0, &dberror);
		if( rc != SQLITE_OK ){
			fprintf(stderr, "SQL error: %s\n", dberror);
			sqlite3_close(db);
			return 0;
		}
	}

	sqlite3_close(db);

	if(!change_rights(path, uid, gid,0))
		return 0;
	
	dir = strrchr(path, '/');

	if(!dir)
	{
		fprintf(stderr, "Unexpected error while changing the rights of the database directory.\n"
				"Please delete the database file and restart aof-db -c.\n");
		return 0;
	}

	*dir = 0;

	if(!change_rights(path, uid, gid,1))
	{
		*dir = '/';
		return 0;
	}

	*dir = '/';

	return 1;
}
