/*
 * filename: src/utils.c
 * author: Pablo Yanez Trujillo <yanezp@informatik.uni-freiburg.de>
 */

#include <sstrings.h> /* string.h, ctype.h stdlib.h included */
#include <stdio.h>
#include <readline/readline.h>
#include <readline/history.h>

#include "utils.h"

SLL* createSLL()
{
	SLL* tmp = NULL;

	tmp = malloc(sizeof(SLL));

	if(tmp)
	{
		tmp->data = NULL;
		tmp->allocated = 0;
		tmp->insert = 0;
		tmp->addtimes = 0;
		tmp->next = NULL;
		tmp->last = NULL;
	}

	return tmp;
}

int addSLL(SLL* sll, void* data, int allocated)
{
	SLL* tmp;
	if(!sll) return 0;

	if(!sll->insert)
	{
		/* initial list */
		sll->data = data;
		sll->allocated = allocated;
		sll->next = NULL;
		sll->last = sll;
		sll->insert = 1;
		sll->addtimes = 1;
		return 1;
	}

	tmp = createSLL();

	if(!tmp)
	{
		fprintf(stderr, "Not enough memory\n");
		return 0;
	}

	tmp->data = data;
	tmp->allocated = allocated;
	tmp->insert = 1;
	sll->last->next = tmp;
	sll->last = tmp;
	sll->addtimes++;
	
	return 1;
}

/* if data == NULL I won't know whether there
 * was a failure or not. That's where `got` is
 * useful */
void* getSLL(const SLL* sll, size_t pos, int* got)
{
	const SLL* tmp;
	size_t count = pos;
	if(!sll)
	{
		*got = 0;
		return NULL;
	}

	tmp = sll;

	while(tmp->next && count)
	{
		tmp = tmp->next;
		count--;
	}

	if(count)
	{
		*got = 0;
		return NULL;
	}

	*got = 1;

	return tmp->data;
}

void cleanupSLL(SLL* sll)
{
	SLL* tmp = sll;
	if(!sll) return;

	/* perhaps is nothing inserted */
	if(!sll->insert)
	{
		free(sll);
		return;
	}

	
	while(tmp)
	{
		tmp = tmp->next;
		if(sll->allocated)
			free(sll->data);

		free(sll);
		sll = tmp;
	}
}

char* readfrom(FILE* buffer, const char* filename, int stop, int* ferr, int* eof)
{
	char* tmp=NULL, *read=NULL;
	int c;
	
	if(!buffer && !filename) return NULL;
	if(!stop) return NULL;
	if(buffer && filename) return NULL;

	if(filename)
	{
		buffer = fopen(filename, "r");

		if(!buffer)
		{
			fprintf(stderr, "Cannot open file `%s'\n", filename);
			*ferr = 1;
			return NULL;
		}
	}

	while( ((c = fgetc(buffer)) != stop) && c != EOF)
	{
		tmp = addchar(&read, c);

		if(!tmp)
		{
			free(read);
			if(filename)
				fclose(buffer);
			fprintf(stderr, "Not enough memory\n");
			return NULL;
		}

		read = tmp;
	}

	if(ferror(buffer)) *ferr = 1;

	if(c == EOF) *eof = 1;

	if(filename)
		fclose(buffer);

	return read;
}

SLL* userinput(const char* questionlist[], int quesnumb, SLL* args)
{
	const char* notice = "To end the input type ENTER and then press Ctrl+D\n";
	char* read = NULL, *line = NULL, *tmp, *tmp2;
	SLL* list = NULL;
	int got, stop, ferr;
	size_t i;

	for(i = 0; i < quesnumb; ++i)
	{
		read = getSLL(args, i, &got);

		if(!got || !read)
		{
			stop = strcmp(questionlist[i+quesnumb], "") ? EOF : '\n';
			got = 0;
			ferr = 0;

			if(strcmp(questionlist[i+quesnumb], "r"))
			{
			
				printf("%s%s", stop==EOF ? notice : "", questionlist[i]);
				line = readfrom(stdin, NULL, stop, &ferr, &got);

				if(!line)
				{
					if(!ferr)
					{
						line = str2sstr("");
	
						if(!line)
						{
							fprintf(stderr, "Not enough memory\n");
							cleanupSLL(list);
							return NULL;
						}
					} else {
						fprintf(stderr, "Could not read from standard input.\n");
						cleanupSLL(list);
						return NULL;
					}
				}
			} else {
				line = readline(questionlist[i]); /* module must free this variable */

				if(!line)
				{
					fprintf(stderr, "Could not read from stdin. Maybe ... not enough memory?\n");
					cleanupSLL(list);
					return NULL;
				}

			}

			if(!list)
			{
				list = createSLL();
				if(!list)
				{
					free(line);
					fprintf(stderr, "Not enough memoy\n");
					return NULL;
				}
			}

			tmp = trim(line);

			if(tmp)
			{
				tmp2 = str2sstr(tmp);

				if(!tmp2)
				{
					fprintf(stderr, "Not enough memoy\n");
					free(line);
					cleanupSLL(list);
					return NULL;
				}

				free(line);
				line = tmp2;
			}


			if(!addSLL(list, (char*) line, 1))
			{
				free(line);
				fprintf(stderr, "Not enough memoy\n");
				cleanupSLL(list);
				return NULL;
			}
		} else {
			if(!list)
			{
				list = createSLL();
				if(!list)
				{
					fprintf(stderr, "Not enough memoy\n");
					return NULL;
				}
			}

			tmp = str2sstr(read);

			if(!tmp)
			{
				fprintf(stderr, "Not enough memoy\n");
				cleanupSLL(list);
				return NULL;
			}

			read = tmp;

			tmp = trim(read);

			if(tmp)
			{
				tmp2 = str2sstr(tmp);

				if(!tmp2)
                                {
                                        fprintf(stderr, "Not enough memoy\n");
                                        cleanupSLL(list);
                                        return NULL;
                                }
				free(read);
                                read = tmp2;
			}

			if(!addSLL(list, (char*) read, 1))
                        {
                                free(read);
                                fprintf(stderr, "Not enough memoy\n");
                                cleanupSLL(list);
                                return NULL;
                        }
		}
	}

	return list;
}

int checknum(const char* buffer, int* num)
{
	int i, len;

	len = strlen(buffer);

	for(i = 0; i < len; ++i)
	{
		if(!isdigit(buffer[i]))
			return 0;
	}

	*num = atoi(buffer);

	return 1;
}

sqlite3* db_connect(const char* path)
{
	FILE *f;
	sqlite3* db;
	int rc;

	f = fopen(path, "r");

	if(!f)
	{
		fprintf(stderr, "The database was not installed yet. Execute aof-db -c\n");
		return NULL;
	}

	fclose(f);

	rc = sqlite3_open(path, &db);

	if(rc)
	{
		fprintf(stderr, "Can't open database: %s\n", sqlite3_errmsg(db));
		sqlite3_close(db);
		return NULL;
	}

	return db;
}

char* prepare_input(const char* input)
{
	/* prepare characters! 
	 * ' --> ' ', 
	 * \ --> ' ',
	 * ! __> ' '
	 *
	 * learn how to escape for sqlite then rewrite this function.
	 */

	char* map, *res = NULL, *count; /*, *tmp; */

	res = str2sstr(input);

	if(!res) return NULL;

	count = res;

	while((map = strstr(count, "\\")))
	{
		*map = ' ';
		/*
		tmp = insertstring(&res, " ", map - res + 1);

		if(!tmp)
			return NULL;

		count = map + 2; 
		*/
		count = map + 1;
	}

	count = res;

        while((map = strstr(count, "'")))
        {
		*map = ' ';
		/*
                tmp = insertstring(&res, " ", map - res + 1);

                if(!tmp)
                        return NULL;

                count = map + 2;
		*/
		count = map + 1;
        }

	count = res;

        while((map = strstr(count, "!")))
        {
		*map = ' ';
		/*
                tmp = insertstring(&res, " ", map - res + 1);

                if(!tmp)
                        return NULL;

                count = map + 2;
		*/
		count = map + 1;
        }

	return res;
}
